// ****************************************************************************
//             Copyright (C) Oasys Design Systems, Inc. 2004 - 2008
//                             All rights reserved.
// ****************************************************************************
#include "LLoad.h"
// nl
#include "NDb"
// utils
#include "UStatCpu"
// system
#include <stdio.h>
#include <sys/types.h>
#include <time.h>
#include <stdlib.h>

#include "LibLoadMsg.h"
#include "LibParam.h"

LCell::~LCell()
{
  delete [] d_pins; delete [] d_order; 
  delete[] d_pinNames; delete [] d_nextStateType;
  for (ModeDefIter iter = d_modeDefs.beginRandom(); iter; ++iter) {
    delete iter.value();
  }
}

LCell::ModeVals*
LCell::addMode(const UString& mode, const UTrace& t)
{
  ModeVals* modeVals;
  ModeDefIter modeDefIter = d_modeDefs.find(mode); if (modeDefIter) {
    LIBLOADMSGT(modeAlreadyExists, t, mode.str());
    modeVals = modeDefIter.value();
  } else {
    modeVals = new IN_CURRENT_POOL ModeVals;
    d_modeDefs.add(mode).value(modeVals);
  }
  return modeVals;
}

BDD
LCell::findModeValue(const UString& mode, const UString& modeVal, const UTrace& t)
{
  ModeDefIter modeDefIter = d_modeDefs.find(mode);
  if (modeDefIter) {
    ModeVals* modeVals = modeDefIter.value();
    ModeValIter viter = modeVals->find(modeVal);
    if (viter) {
      return viter.value();
    } else {
      LIBLOADMSGT(modeValueNotDefined, t, modeVal.str(), mode.str());
    }
  } else {
    LIBLOADMSGT(modeNotDefined, t, mode.str());
  }
  return BDD();
}

bool
LCell::addModeValue(ModeVals* modeVals, const UString& modeVal, BDD f)
{
  ModeValIter viter = modeVals->find(modeVal);
  if (viter) {
    return false;
  } else {
    modeVals->quickAdd(modeVal).value(f);
  }
  return true;
}

NLib
NDb::readLibrary(const char* filename)
{
  UStatCpu::enter(USC_Lib);
  UTask* key = control().lock("Reading Timing Library", 0);
  LLoad ll = LLoad(filename, *this, false, false, false);
  if (ll.library()) {
    ll.library().postProcess();
  }
  delete key;
  UStatCpu::leave(USC_Lib);
  return ll.library();
}

LLoad::LLoad(const char* filename, NDb db, bool writeOut, bool printStats,
	     bool warn) :
  d_libCorner(TLibCorner::maximum),
  d_warnNoSupport(warn)
{
  extern char* curr_file;
  LibErrorT err;
  time_t startt, endt;

#ifdef LIBERTY_DEBUG
  LibParse::debugMode(1);
#endif
  LibParse::init(db.control());
  uprintf("\n\nreading %s...\n", filename);
  time(&startt);
  curr_file = const_cast<char*>(filename);
  err = LibParse::read(filename);
  if (err == LIB_INVALID_NAME) {
    uprintf("COULD NOT OPEN %s for parsing-- quitting...\n",filename);
    LibParse::quit();
    return;
  } else if (err == LIB_SYNTAX_ERROR && !UParam::readLibContinueOnError()) {
    uprintf("\nSyntax Errors were detected in the input file!\n\n");
    LibParse::quit();
    return;
  }
  time(&endt);
  uprintf("Finished reading. \nElapsed time= %ld seconds\n", endt-startt);
  parseTree2Db(db);

  if (writeOut) {
    LibGroupIter giter;
    LibGroup* group;
    LibGroup* masterGroup = LibParse::group();
    char buf1[100];
    int count = 1;

    giter = masterGroup->groupIter();
    while ((group = giter++)) {
      sprintf(buf1,"%s.blatantcopy%d", group->firstName().str(), count++);
      uprintf("writing %s...\n", buf1);
      time(&startt);
      LibParse::write(buf1, group);
      time(&endt);
      uprintf("... Done. Elapsed time= %ld seconds\n", endt-startt);
    }
  }
	
  if (printStats) {
    LibGroupIter giter;
    LibGroup* group;
    LibGroup* masterGroup = LibParse::group();
    int cellcount=0;

    uprintf("\n\nStats:\n");
    giter = masterGroup->groupIter();
    while ((group = giter++)) {
      const char* gname = group->firstName();
      LibGroupIter giter2;
      LibGroup* g2;
      if( !gname )
	gname = "<NONAME>";
      uprintf("\n\nLibrary %s:\n", gname);
      giter2 = group->groupIter();
      while ((g2 = giter2++)) {
	const char* gt = g2->type();
	if( !strcmp(gt,"cell") )
	  cellcount++;
      }
      uprintf("   cells: %d\n\n", cellcount);
    }
  }

  uprintf ("BDD Manager Var Count = %d\n", db.bddMgr().ReadSize());
  LibParse::quit();
  return;
}

void
LLoad::addLib(const LibGroup* llib, NDb db)
{
  d_lib = db.findLibrary(llib->firstName());
  if (d_lib) {
    LIBLOADMSG(libraryAlreadyLoaded, llib->firstName().str());
  } else {
    d_lib = db.newLibrary(llib->firstName(), llib->filename());
    // process library attributes
    processLibAttr(llib);
  }
  // process library subgroups
  processLibSubgroup(llib, db);
}

void
LLoad::parseTree2Db(NDb db)
{
  LibErrorT err;
  LibGroupIter giter;
  LibGroup* llib;
  LibGroup* masterGroup;
  time_t startt,endt;

  uprintf("Loading library into database...\n");
  time(&startt);
  masterGroup = LibParse::group();
  giter = masterGroup->groupIter();
  while ((llib = giter++)) {
    err = LibParse::checkSyntax(llib);
    if (!db.control().interrupted()) {
      if (err != LIB_NO_ERROR && !UParam::readLibContinueOnError()) {
	uprintf("Errors detected\nLibrary %s not loaded.\n",
		llib->firstName().str());
      } else {
	addLib(llib, db);
      }
    }
  }
  time(&endt);
  uprintf("Finished loading library. \nElapsed time= %ld seconds\n", endt-startt);
}

// when called during input ordering:
//    idxOffset = bidirPinCount;
//    slice.offset() = outPinCount+bidirPinCount
void
LCell::permutePins(UAggSlice slice, int permute[], int idxOffset)
{
  int      count   = slice.count();
  int      from    = slice.offset();
  TLibPin* tmp     = new IN_CURRENT_POOL TLibPin[count];
  UAggIter iter;
  int      i;

  for (iter = slice.msbToLsb(); iter; iter++) {
    int idx = iter.current() - from;
    tmp[idx].grab(d_pins[iter.current()]);
  }
  for (iter = slice.msbToLsb(), i = 0; iter; iter++, i++) {
    int idx = permute[i] - idxOffset;
    uassert (idx >= 0 && idx < count);
    d_pins[iter.current()].grab(tmp[idx]);
    d_order[iter.current()] = from + idx;
  }
  
  delete [] tmp;
}
    
    
